/* lqbyteline.c -- Copyright 1993 Liam R. E. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 *
 * $Id: lqbyteline.c,v 1.2 93/12/08 23:14:14 lee Exp $
 *
 * Turns byte offsets into line numbers; see lqkwic -L, and also the
 * example use in the lq.sh shell script.
 */

#include "error.h"
#include "globals.h" /* defines and declarations for database filenames */
#include "readfile.h"
#include "emalloc.h"

#include <stdio.h>

#include <sys/types.h>

extern int AsciiTrace;

/** System calls and functions... **/
/** Unix system calls used in this file: **/
extern void exit();

/** Unix Library Functions used: **/
/** lqtext library functions: **/
extern void SetDefaults();
extern void DefaultUsage();

/** functions used before they're defined within this file: **/
static void ProcessFile();
static void PrintLineNumber(/*Offset, FileName*/);

/** **/

static char *Revision = "@(#) $$";

char *progname = "lqbyteline";

int
main(argc, argv)
    int argc;
    char *argv[];
{
    extern int optind, getopt();
    extern char *optarg;
    int ch;
    int ErrorFlag = 0;
    char *InputFile = 0;

    progname = argv[0];

    SetDefaults(argc, argv);

    while ((ch = getopt(argc, argv, "Zz:af:hNpr:slxVv")) != EOF) {
	switch (ch) {
	case 'z':
	case 'Z':
	    break; /* done by SetDefaults(); */
	case 'V':
	    fprintf(stderr, "%s version %s\n", progname, Revision);
	    break;
	case 'v': /* same as -t 1 */
	    AsciiTrace++;
	    break;
	case 'f':
	    InputFile = optarg;
	    break;
	case 'x':
	    ErrorFlag = (-1);
	    break;
	case '?':
	    ErrorFlag = 1;
	}
    }

    if (ErrorFlag) {
	fprintf(stderr, "Usage: %s [options] [file [...]]\n", progname);
	DefaultUsage();
	exit( ErrorFlag > 0 ? 1 : 0); /* 0 means -x was used */
    }
    
    if (InputFile) {
	ProcessFile(InputFile);
    }

    if (optind < argc) {
	ProcessFile("-");
    } else {
	while (optind < argc) {
	    ProcessFile(argv[optind++]);
	}
    }

    return 0;
}

static void
ProcessFile(InputFile)
    char *InputFile;
{
    FILE *f;
    char NeedClose = 1;
    char *theLine;

    if (STREQ(InputFile, "-")) {
	f = stdin;
	NeedClose = 0;
	InputFile = "standard input";
    } else {
	extern FILE *fEopen();

	f = fEopen(E_FATAL|E_SYS, InputFile, "List of bytes and files", "r");
    }

    while (fReadLine(f, &theLine, UF_NORMAL) >= 0) {
	if (theLine && *theLine) {
	    unsigned long theOffset;
	    register char *p;

	    for (p = theLine; *p; p++) {
		if (!isspace(*p)) break;
	    }

	    if (!isdigit(*p)) {
		Error(E_FATAL,
		    "\"%s\" contains line with no leading byte offset: %s",
		    InputFile
		);
		(void) efree(theLine);
		break;
	    }

	    theOffset = 0L;

	    do {
		theOffset *= 10;
		theOffset += (*p) - '0';
		p++;
	    } while (isdigit(*p));

	    if (!isspace(*p)) {
		Error(E_FATAL,
		    "\"%s\" no space and filename after offset in: %s",
		    InputFile,
		    theLine
		);
		break;
	    }
	    *p = '\0';

	    do {
		++p;
	    } while (isspace(*p));

	    PrintLineNumber(theOffset, p);
	}
	if (theLine) {
	    (void) efree(theLine);
	}
    }

    if (NeedClose) {
	(void) close(f);
    }
}

static void
PrintLineNumber(Offset, FileName)
    unsigned long Offset;
    char *FileName;
{
    static FILE *f = 0;
    static unsigned long LastOffset = 0L;
    static unsigned long LastLineNumber;
    static char *LastName = 0;
    int ch;

    if (!LastName || !STREQ(FileName, LastName) || LastOffset > Offset) {
	if (f) {
	    (void) fclose(f);
	    f = 0;
	}
	if (LastName) {
	    efree(LastName);
	}
	LastName = emalloc("PrintLineNumber::FileName", strlen(FileName) + 1);
	(void) strcpy(LastName, FileName);
	f = fEopen(E_WARN, LastName, "input file", "r");
	if (!f) {
	    return;
	}
	LastOffset = 0L;
	LastLineNumber = 1L;
    }

    while (LastOffset != Offset) {
	if ((ch = getc(f)) == EOF) {
	    break;
	} else if (ch == '\n') {
	    ++LastLineNumber;
	}
	++LastOffset;
    }

    printf("%ld %s\n", LastLineNumber, FileName);
}
